__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

import numpy as np
np.random.seed(12349)


from sfctools.examples.example_wrapper import Example

from sfctools.datastructs.bank_order_book import BankOrderBook
from sfctools.bottomup.matching import MarketMatching
from sfctools import Agent
import matplotlib.pyplot as plt


def run(show_plot=False):

    # create a bank agent
    class Bank(Agent):
        def __init__(self):
            super().__init__()

    class MyCreditMarket(MarketMatching):
        def __init__(self):
            super().__init__()

        def rematch(self):
            # link some agents

            r_l = 0.05 # interest on loans

            self.link_agents(self.supply_list[0],self.demand_list[1],r_l)
            self.link_agents(self.supply_list[0],self.demand_list[5],r_l)
            self.link_agents(self.supply_list[3],self.demand_list[5],r_l)

            new_loans = 4.2 # quantity
            duration = 5    # 5 periods duration

            for bank in [self.supply_list[0],self.supply_list[3]]:
                for creditor in [self.demand_list[1],self.demand_list[5]]:
                    bank.order_book.add_loans(creditor,new_loans,r_l,t=duration)

    class MyDepositMarket(MarketMatching):
        def __init__(self):
            super().__init__()

        def rematch(self):
            # link some agents

            r_d = 0.03 # interest on deposits all same

            self.link_agents(self.supply_list[2],self.demand_list[3],r_d)
            self.link_agents(self.supply_list[2],self.demand_list[6],r_d)
            self.link_agents(self.supply_list[4],self.demand_list[6],r_d)

            new_deposits = 1000.0 # quantity of deposits all the same

            # add deposits at banks
            for bank in [self.demand_list[3], self.demand_list[6]]:
                for depositor in [self.supply_list[2],self.supply_list[4]]:
                    bank.order_book.add_deposits(depositor,new_deposits, r_d)

    # create some agents
    firms = [Agent() for i in range(12)] # create some firms seeking for creidt
    depositors = [Agent() for i in range(25)] # create some depositors
    banks = [Bank() for i in  range(10)] # create banks

    # create markets
    credit_market = MyCreditMarket()
    deposit_market = MyDepositMarket()

    # register agents at markets
    for f in firms:
        credit_market.add_demander(f)

    for d in depositors:
        deposit_market.add_supplier(d)

    for b in banks:
        credit_market.add_supplier(b)
        deposit_market.add_demander(b)

    for b in banks: # add order book to banks
        b.order_book = BankOrderBook(b,credit_market,deposit_market)

    # match the markets
    deposit_market.rematch()
    credit_market.rematch()

    # look at some output
    df_depos,df_loans = banks[3].order_book.to_dataframe()

    print(df_depos, "\n")
    # print(df_depos["Agent"])

    loans = banks[3].order_book.get_loans_of(firms[5])

    # process the loan
    loans[0]["Time"] += 1

    # this is an example
    # insert your code here...
    t_remain = loans[0]["Maturity"] - loans[0]["Time"]
    loans[0]["Outstanding Loans"] -= 0.1

    # check in original data
    df_depos,df_loans = banks[3].order_book.to_dataframe()

    print(df_loans, "\n")
    # print(df_loans["Time"])

    # plot the resulting market connections
    if show_plot:
        deposit_market.plot()
        credit_market.plot()


class BankOrderBookExample(Example):
    def __init__(self, show_plot=False):
        super().__init__(lambda: run(show_plot=show_plot))


if __name__ == "__main__":

    my_ob = BankOrderBookExample(show_plot=True)
    my_ob.run()



from PyQt5 import QtGui
import os
import pickle
from PyQt5.QtWidgets import QFileDialog,QMessageBox

class ThemeManager:
    """
    the theme manager takes care of the theme
    """

    def __init__(self,parent):
        self.theme = "bright" # current state

        self.parent = parent
        self.restore()
        self.christmas_mode = False 

        self.style_dark = {
            "main": self.read_file("styles/dark/main.txt")
            }

        self.style_bright ={
            "main": self.read_file("styles/bright/main.txt")
        }

        self.styles  = {
        "bright": self.style_bright,
        "dark": self.style_dark
        }

    def activate_christmas_mode(self):
        print("*** CHRISTMAS MODE ***")
        self.christmas_mode = True 

        self.style_bright ={
            "main": self.read_file("styles/christmas/main.txt")
        }

        self.styles  = {
        "bright": self.style_bright,
        "dark": self.style_dark
        }

        self.parent.switch_theme()
        self.parent.switch_theme()
        

    def restore_buttons(self):
        btn_table = {
                1: self.parent.bc1,
                2: self.parent.bc2,
                3: self.parent.bc3,
                4: self.parent.bc4,
                5: self.parent.bc5,
                6: self.parent.bc6,
                7: self.parent.bc7,
                8: self.parent.bc8,
            }

        for i, color in enumerate(self.colors[self.theme]):
            btn_table[i+1].setStyleSheet("background-color: rgb(%s,%s,%s);"%(color.red(),color.green(),color.blue()))
    
    def restore(self):
        # default values
        self.dark_theme = [QtGui.QColor(50, 50, 50 , 255),
                           QtGui.QColor(120, 120, 120, 255),
                           QtGui.QColor(220,220,220, 120), # transaction lines
                            QtGui.QColor(177,177,177 , 255),
                            QtGui.QColor(175,175,175, 255),
                            QtGui.QColor(180, 180, 180, 255),
                            QtGui.QColor(72,72,72 , 255), # box labels
                            QtGui.QColor(120,120,120,80) # background
                           ]

        self.bright_theme =[QtGui.QColor(50, 50, 50 , 255),
                           QtGui.QColor(20, 20, 20, 255),
                           QtGui.QColor(81, 81, 81 , 80), # transaction lines
                            QtGui.QColor(221,221,221, 200),
                           QtGui.QColor(255,96,99,255), # QtGui.QColor(230, 230, 230, 255), # active connection
                            QtGui.QColor(210, 210, 210, 255),
                            QtGui.QColor(47, 47,47 , 255), # box labels
                            QtGui.QColor(80,80,80,100) # background
                           ]

        self.colors = {
            "dark": self.dark_theme,
            "bright": self.bright_theme
        }
        self.restore_buttons()
        self.parent.update()


    def load_colors(self, filename=None):

        self.restore()

        if filename is None:
            filename = QFileDialog.getOpenFileName(self.parent, 'Open file',os.getcwd(), "Attune Theme Files (*.sfctheme)")[0]

        if filename is not None and filename != "":

            success = False
            try:
                with open(filename,"rb") as file:
                    theme = pickle.load(file)
                    self.colors[self.theme] = theme
                    self.restore_buttons()
                success = True
            except:
                pass

            if not success:
                try: # try with relative path ?
                    filename = os.path.split(filename)[1]
                    foldername = os.path.dirname(self.parent.current_file)
                    filename = os.path.join(foldername,filename)
                    with open(filename,"rb") as file:
                        theme = pickle.load(file)
                        self.colors[self.theme] = theme
                        self.restore_buttons()
                    success = True
                except:
                    pass

            if not success:
                self.parent.notify("Could not open theme from %s" % filename , title="Error")

    def save_colors(self,filename=None):
        # save the colortheme to reative path

        if filename is None:
            return

        try:
            # get the path of the main file
            foldername = os.path.dirname(self.parent.current_file)

            # join path to save under same folder as main file
            filename = os.path.join(foldername,filename)

            btn_table = {
                    1: self.parent.bc1,
                    2: self.parent.bc2,
                    3: self.parent.bc3,
                    4: self.parent.bc4,
                    5: self.parent.bc5,
                    6: self.parent.bc6,
                    7: self.parent.bc7,
                    8: self.parent.bc8,
                }

            ssheets = []
            for i,color in enumerate(self.colors[self.theme]):
                clr = btn_table[i+1].palette().button().color()
                clr.setAlpha(color.alpha())
                ssheets.append(clr)
                print("save theme", clr.alpha())

            print(ssheets)

            if filename is None:
                filename = QFileDialog.getSaveFileName(self.parent, 'Save file',os.getcwd(), "Attune Theme Files (*.sfctheme)")[0]


            if filename is not None and filename != "":
                with open(filename,"wb") as file:
                    pickle.dump(ssheets,file)

        except:
            self.parent.notify("An error occurred.",title="Error")

    def get_color(self,category):
        """
        get the color given the current theme
        :param category: int, which layer of the color theme you want
        """
        return self.colors[self.theme][category]

    def get_stylesheet(self,category):
        return self.styles[self.theme][category]


    def get_notification_style(self):
        # style of notifiaction window

        notification_styles = {
            "dark": "QLabel{color:white}",
            "bright": "QLabel{color:black}"
        }

        return notification_styles[self.theme]

    def get_table_style(self):

        dark_table_style = self.read_file("styles/dark/tables.txt")
        bright_table_style = self.read_file("styles/bright/tables.txt")

        if self.christmas_mode:
            bright_table_style = self.read_file("styles/christmas/tables.txt")


        table_styles = {
            "dark": dark_table_style,
            "bright": bright_table_style
        }

        return table_styles[self.theme]


    def get_background_style(self):
        # background color for dialogs

        dark_bg_style=self.read_file("styles/dark/background.txt")
        bright_bg_style=self.read_file("styles/bright/background.txt")

        if self.christmas_mode:
            bright_table_style = self.read_file("styles/christmas/background.txt")

        notification_styles = {
            "dark": dark_bg_style,
            "bright": bright_bg_style
        }

        return notification_styles[self.theme]



    def read_file(self,fname):

        # fname = "./attune/src/styles/" + fname
        home = os.path.dirname(os.path.abspath(__file__))
        fname = os.path.join(home, fname)

        with open(fname,"r") as file:
            return str(file.read())



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'test' # options are: dev, test, prod

from .treestruct import TreeStruct
from ..core.agent import Agent

from collections import defaultdict

import networkx as nx
import matplotlib.pyplot as plt
from networkx.drawing.layout import spring_layout


def dict_depth(d):
    """
    helper function to get the depth of a nested dict
    """
    if isinstance(d, dict):
        return 1 + (max(map(dict_depth, d.values())) if d else 0)
    return 0


def unfold_tree(tree, agent_name_dict):
    """
    iterates through production tree and 'unfolds it' by inserting the appropriate agent names
    """

    t = {}  # tree to store production chain

    # if isinstance(tree,list):
    #    strtree += "|        "*rec + str( tree) + "\n"
    #    return tree,strtree


    for key, val in tree.items():
        if val == {}: # is a leaf?

            if (not isinstance(agent_name_dict,defaultdict)) and key.split(";")[0] not in agent_name_dict:
                error =  KeyError("Error building production tree: could no find definition for leaf %s" % key.split(";")[0])
                raise error

            name = agent_name_dict[key.split(";")[0]]  # name is split from production function definition here
            if isinstance(key, str) and len(key.split(";")) > 1:
                display_name = name + ";" + key.split(";")[1]

            else:
                if isinstance(name,list): # <- can pass a one-element list or just an instance here
                    display_name = str(name[0])
                else:
                    display_name = str(name)

            t[display_name] = {}  # str(key)  # is a leaf

        else: # is no leaf?

            if (not isinstance(agent_name_dict,defaultdict)) and key.split(";")[0] not in agent_name_dict:
                error =  KeyError("Error building production tree: could no find definition for node %s" % key.split(";")[0])
                raise error

            for name in agent_name_dict[key.split(";")[0]]:

                if isinstance(name, str) and len(key.split(";")) > 1:
                    display_name = name + ";" + key.split(";")[1]
                else:
                    if isinstance(name,list): # <- can pass a one-element list or just an instance here
                        display_name = str(name[0])
                    else:
                        display_name = str(name)

                # print("strtree",strtree)
                t[display_name] = unfold_tree(val, agent_name_dict)

    return t

def tree_to_coords(tree, rec=0, data=None,linedata=None, style="tree"):
    """
    makes plotting data for a tree

    :param tree: a dict of dicts
    :param style: 'tree' or 'vertical' or 'horizontal'

    :param data: recursion parameter, do not change
    :param rec: recursion parameter, recursion depth, do not change

    :return strtree: str, tree in string representation

    """
    dx = 30

    if data is None:
        data = []

    if linedata is None:
        linedata = []


    if tree is not None:
        for key, val in tree.items():

            if style == "vertical":
                # strtree += "|        " * rec + str(key) + "\n"
                #x += dx*rec
                #y  += dy
                #data.append((str(key),x,y,False))
                raise NotImplementedError("Not implemented. try 'tree'")
            elif style == "horizontal":
                # strtree += "|" + "____________" * rec + str(key) + "\n"
                #x += dx*rec
                #y += dy
                #data.append((str(key),x,y,False))
                raise NotImplementedError("Not implemented. try 'tree'")

            elif style == "tree":
                if rec >= 1:
                    if val == {}: # is leaf
                        # strtree += "'            " * (rec - 1) + "|" + "____________. " + str(key) + "\n"
                        x = dx*rec
                        data.append((str(key),x,rec,True))
                        # linedata.append((x,rec-5,x,rec))
                        linedata.append((x-dx,rec,x,rec))

                    else:
                        # strtree += "'            " * (rec - 1) + "|" + "____________ " + str(key) + "\n"
                        x = dx*rec
                        data.append((str(key),x,rec,False))
                        #linedata.append((x,rec-5,x,rec))
                        linedata.append((x-dx,rec,x,rec))

                else:
                    # strtree += "|" + str(key) + "\n"
                    x = 0
                    # linedata.append((x,rec-5,x,rec))
                    data.append((str(key),x,rec,False))


            tree_to_coords(val, rec + 1, data, linedata,style)

    return data,linedata



def tree_to_str(tree, rec=0, strtree="", style="tree"):
    """
    prints a tree structure (nested dict) as str

    :param tree: a dict of dicts
    :param style: 'tree' or 'vertical' or 'horizontal'

    :param strtree: recursion parameter, do not change
    :param rec: recursion parameter, recursion depth, do not change

    :return strtree: str, tree in string representation

    """

    if tree is not None:
        for key, val in tree.items():

            if style == "vertical":
                strtree += "|        " * rec + str(key) + "\n"

            elif style == "horizontal":
                strtree += "|" + "____________" * rec + str(key) + "\n"


            elif style == "tree":
                if rec >= 1:
                    if val == {}:
                        strtree += "'            " * (rec - 1) + "|" + "____________. " + str(key) + "\n"
                    else:
                        strtree += "'            " * (rec - 1) + "|" + "____________ " + str(key) + "\n"
                else:
                    strtree += "|" + str(key) + "\n"
            strtree = tree_to_str(val, rec + 1, strtree, style)

    return strtree


def recursive_items(dictionary):
    # yields items of a dictionary recursively
    for key, value in dictionary.items():
        if type(value) is dict:
            yield key.split(";")  # (key, value)
            yield from recursive_items(value)
        else:
            yield key.split(";")  # (key, value)


def reverse_dict_search(d, x):
    # search key x in values of dictionary d
    for k, v in d.items():
        #print("search", k, v)
        if x == v or x in v:
            return k


def extract_production_info(tree, production_dict, agent_name_dict):
    """
    extracts relevant production parameters from production_dict and agent_name_dict,
    see ProductionTree class for more information.
    """

    #print("EXTRACT")
    keys = list(recursive_items(tree))

    production_info = {}

    #p#rint("production info", keys)

    for key in keys:

        factor = key[0]
        agents = agent_name_dict[factor]

        params = None
        if len(key) > 1:
            params = production_dict[key[1]].copy() # copy ensures that product key does not change 'globally'
            params["product"] = factor

        if isinstance(agents, list):
            for agent in agents:
                if agent not in production_info:
                    production_info[agent] = params
        else:
            if agents not in production_info:
                production_info[agents] = params

        # print("agents",agents, "---->",factor)

    return production_info


def full_tree_pos(G):
    # position generator from https://stackoverflow.com/questions/33439810/preserving-the-left-and-right-child-while-printing-python-graphs-using-networkx

    n = G.number_of_nodes()
    if n == 0 : return {}
    # Set position of root
    pos = {0:(0.5,0.9)}
    if n == 1:
        return pos
    # Calculate height of tree
    i = 1
    while(True):
        if n >= 2**i and n<2**(i+1):
            height = i
            break
        i+=1
    # compute positions for children in a breadth first manner
    p_key = 0
    p_y = 0.9
    p_x = 0.5
    l_child = True # To indicate the next child to be drawn is a left one, if false it is the right child
    for i in range(height):
        for j in range(2**(i+1)):
            if 2**(i+1)+j-1 < n:
                # print 2**(i+1)+j-1
                if l_child == True:
                    pos[2**(i+1)+j-1] = (p_x - 0.2/(i*i+1) ,p_y - 0.1)
                    G.add_edge(2**(i+1)+j-1,p_key)
                    l_child = False
                else:
                    pos[2**(i+1)+j-1] = (p_x + 0.2/(i*i+1) ,p_y - 0.1)
                    l_child = True
                    G.add_edge(2**(i+1)+j-1,p_key)
                    p_key += 1
                    (p_x,p_y) = pos[p_key]

    return pos

class Dummy:
    def __init__(self):
        pass


class ProductionTree:
    """
    A production tree is a set of nested dictionaries to describe a production structure (or supply chain).



        Example:

        .. code-block:: python

            from sfctools import Agent, ProductionTree

            class MyProduction():
                def __init__(self,x):


            my_agent_a = MyProduction("A")
            my_agent_b = MyProdution("B")
            my_agent_y = MyProduction("Y")

            production_dict = {
                    "CES_1": {"epsilon": 1.0, "alpha": [0.5,0.3,0.2]}, #...
                    }

            data =  {"Y;CES_1":
                        {"A": {},
                         "B": {}}

            agent_name_dict = {
                            "A": [my_agent_a],
                            "B": [my_agent_b],
                            "Y": [my_agent_y]
                        }

            my_production_tree = ProductionTree(production_dict,data,agent_name_dict)
    """

    count = 0

    def __init__(self, production_dict, data, agent_name_dict=None):
        """
        constructor for production tree

        :param production_dict: stores production function names (str) as keys, and dicts of productino function parameters as values
        :param data:  a nested structure of goods and their production functions, separated by ';' in the keys (see example)
        :param agent_name_dict: A dictionary where the keys are the goods and the values are lists of agents. If None (default), dummy instances will be placed here
        """

        #TODO maybe a bit more documentation about what is going on here is required. -> provide later

        self.production_dict = production_dict

        self.data = data

        if agent_name_dict is None:
            # empty agent name dict -> place dummies here

            agent_name_dict = defaultdict(lambda: [Dummy()])
            print("Default dict", agent_name_dict)
            print(agent_name_dict["Output"])

        self.detailed_data = unfold_tree(data, agent_name_dict)

        """
        This will create two trees: one with the aggregate production tree and one detailed (unfolded one)
        """

        self.production_info = extract_production_info(data, production_dict, agent_name_dict)
        #print("PRODUCTION INFO")
        #for k,v in self.production_info.items():
        #    print(k,v)

        self._root = next(iter(self.data.keys()))
        self._detailed_root = next(iter(self.detailed_data.keys()))

        self.depth = dict_depth(self.data)
        self.detailed_depth = dict_depth(self.detailed_data)

        self.tree = TreeStruct(data, None)
        self.detailed_tree = TreeStruct(self.detailed_data, None)

        self.__class__.count+= 1

    @property
    def root(self):
        """
        returns the root node of the (unfolded) tree structure.
        """
        return self._detailed_root

    def __repr__(self):
        return "< ProductionTree %i >"  % self.__class__.count

    def to_str(self, mode='aggregate'):
        """
        returns a string representation of the production tree
        :param mode: 'detailed' or 'aggregate'
        """
        if mode == "aggregate":
            return tree_to_str(self.data)

        elif mode == "detailed":
            return tree_to_str(self.detailed_data)

        else:
            raise (KeyError("Can not find str conversion mode. Allowed are 'detailed' or 'aggregate'"))


    def plot(self, mode="aggregate", with_labels = True, param_name="sigma", nodecolor="mediumslateblue",show_plot=True):
        """
        generates a plot of the tree structure
        :param mode: 'detailed' or 'aggregate'
        :param param_name: name of the production function parameter to search for (default 'sigma', i.e. elasticity)
        :param nodecolor: color of the nodes of the tree
        :param show_plot: if True, window opens. if False, figure is returned only

        :return fig: matplotlib figure

        """

        t = None
        d = 0
        root = None

        if mode == "aggregate":

            t = self.tree.tree
            d = self.depth - 1
            root = self._root

        elif mode == "detailed":
            t = self.detailed_tree.tree
            d = self.detailed_depth - 1
            root = self._detailed_root

        else:
            raise (KeyError("Can not find plotting mode. Allowed are 'detailed' or 'aggregate'"))

        # t2 = nx.dfs_tree(t, source=root, depth_limit=d)

        #G = nx.Graph()
        #G.add_nodes_from(t)
        ##pos = full_tree_pos(G)

        #plt.show()

        #plt.figure(figsize=(7, 5))
        #pos = full_tree_pos(t) # spring_layout(t) # , prog='dot')
        #nx.draw(t, pos, node_size=25, with_labels=with_labels, arrows=True, node_color='gray')
        #nx.draw(G, pos=pos, with_labels=True)
        #plt.show()


        """
        new plotting script
        """

        #print(self.data)
        coords,linedata = tree_to_coords(self.data)
        #print(coords)

        ymin = 0
        xmax = 0

        x_vals = []
        y_vals = []

        y = 1

        dxlabel = 2
        n = 0

        old_point = None
        old_point_2 = None
        rec_old = 0

        fig = plt.figure(figsize=(3.5,4.5))

        for line_i in linedata:

            #print("LINE",line_i)

            y -= 1

        y = 0

        for coords_i in coords:
            #print("COORDS",coords_i)

            label, x, rec, isleaf= coords_i

            x_vals.append(x)
            y_vals.append(y)

            plt.annotate(label.split(";")[0],(x+dxlabel,y))


            if len(label.split(";")) > 1:
                params = self.production_dict[label.split(";")[1]]

                if param_name in params:
                    plt.annotate(str(params[param_name]),(x+dxlabel,y-.8),color="gray")
                else:
                    plt.annotate(str(params),(x+dxlabel,y-.8),color="gray")


            xmax = max(xmax,x)
            ymin = min(ymin,y)

            y -= 1

        plt.scatter(x_vals,y_vals,color=nodecolor)

        for i in range(len(x_vals)):

            a = x_vals[i]-30

            while a >= 0:
                plt.plot([x_vals[i]-a-30,x_vals[i]-a-30],[y_vals[i],y_vals[i]+1],"--",color="gray",alpha=.5)

                a-= 30


            if y_vals[i] != 0:
                plt.plot([x_vals[i]-30,x_vals[i]-30],[y_vals[i],y_vals[i]+1],color="black")
                plt.plot([x_vals[i]-30,x_vals[i]],[y_vals[i],y_vals[i]],color="black")




        # plt.ylim([ymin,0])
        # plt.xlim([0,xmax])
        plt.axis("off")
        plt.tight_layout()

        if show_plot:
            plt.show()

        return fig


        # print(root.children)



    def traverse(self, subject, direction="up"):
        """
        traverse the tree struct and execute production nodes
        on the way. -> Breadth-first search

        :param subject: the subject carried with the traversal signal
        :param direction: the direction of the trigger signal (traversal is as in nx.bfs)
        """

        for u, v in self.detailed_tree.tree.edges:  # iterate through all edges and trigger handlers

            if direction == "up":
                v.trigger(subject, u)

            elif direction == "down":
                u.trigger(subject, v)

            else:
                raise ValueError(direction)

    def init_nodes(self, subject):
        """
        send production_dict to the nodes for initialization

        :param subject: subject for initalization trigger
        """

        for agent in self.detailed_tree.tree.nodes:
            upstream_suppliers = [x[1] for x in self.detailed_tree.tree.out_edges(agent)]
            agent.trigger(subject, upstream_suppliers, self.production_info[agent])

    def traverse_nodes(self, subject):
        """
        traverse the tree struct nodes

        :param subject: the subject carried with the traversal signal
        """

        for agent in self.detailed_tree.tree.nodes:
            agent.trigger(subject)



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

from scipy.optimize import minimize,dual_annealing
import subprocess
import numpy as np
from functools import partial
import time
import os
import matplotlib.pyplot as plt


"""
EXPERIMENTAL and not tested!

Calibration suite for a general model. BETA requires further testing. Usage at own risk.
"""

class DistanceMeasure:
    """
    Generic class for distance measure.
    """

    def __init__(self,method):
        """
        param kind: determines the method that is applied to compute the distance measure
        """
        self.method = method
        self.callable_dict = {
            "rmse": self.rmse
        }

        self.logger = [] # logs all distances computed

        self.counter = 0

    def d(self, predictions, targets, params, d_params):
        """
        EXPERIMENTAL and not tested!

        compute the distance

        :param prediction: vector of predictions y~[t]
        :param targets: vector of targets y[t]
        :param params: vector of fit paramters used to generate y~[t]
        :param d_params: dict of parameters for calculation of distance
        """

        self.counter += 1
        if self.counter == 10:
            self.counter = 0
            plt.figure()
            plt.style.use("macro")

            colors = ["black","red","blue","lime","magenta","orange"]
            for i in range(predictions.shape[1]):
                plt.plot(predictions[:,i],linestyle="dotted",color=colors[i%len(colors)],alpha=.8)

            for i in range(targets.shape[1]):
                plt.plot(targets[:,i],linestyle="solid",color=colors[i%len(colors)])

            plt.xlabel("Time")
            plt.ylabel("Value")
            filename = "./figures/" + str(time.time())+ ".png"
            print("filename",filename)
            plt.savefig(filename)
            plt.close()


        if self.method in self.callable_dict:

            callable = self.callable_dict[self.method]
            return callable(predictions,targets,params,d_params)

        else:

            raise NotImplementedError("Could not find any implementation for '%s'" % self.method)



    def rmse(self,predictions, targets, params,d_params):
        """
        EXPERIMENTAL and not tested!

        Root-mean-square error

        d_params {"lambda": ...} regularization parameter
        """

        # compute a distance of dimenion of the number of observables in the data (or simulation)

        # evaluate the disance with a norm
        d = np.linalg.norm(predictions-targets,ord=2)
        p = d_params["lambda"] # 0.001 # regularization term
        l = np.sum(np.power(params,2))  # regularization parameter

        L = d + l*p

        self.logger.append(L)

        return L


class CalibrationRoutine:
    """
    EXPERIMENTAL and not tested!

    Generic class for model calibration
    """

    def __init__(self, simulation_ref, datashape, param_bounds, param_init, method="scipy", mode="callable"):
        """
        Calibration for models outputting several time series

        :param simulation_ref:
                in 'shell' mode:  str, path to a python file to be executed as the model
                in 'callable' mode:

        :param datashape: tuple (m,n), exected dimensions of output data
        :param method: method to use for calibration
        :param param_bounds: parameter boundaries dictionary of the form {str(name): tuple(min,max)}
        :param param_init: initial conditions for parameters
        :param mode: 'shell' or 'callable'
        """

        self.method_dict = {
            "scipy": partial(self.calibrate_scipy,method="SLSQP") # uses scipy.optimize to get the work done
        }

        self.simulation_ref  = simulation_ref
        self.param_bounds = param_bounds
        self.x0 = param_init
        self.datashape = datashape
        self.data_measured = None

        self.method = method # determine a calibration method
        self.dmeasure = DistanceMeasure("rmse") # define a distance measure

        self.mode = mode

    def plot(self):
        plt.figure()
        plt.style.use("macro")
        plt.semilogy(self.dmeasure.logger)
        plt.xlabel("Iteration")
        plt.ylabel("Objective Value")
        plt.title("Learning Curve")
        plt.show()

        self.dmeasure.logger = []

    def iter(self,params):
        """
        EXPERIMENTAL and not tested!

        run one iteration of the model
        :param params: a list or arrays of parameters as input to the model
        """

        if self.mode == "shell":

            start = time.time()
            args = " ".join(list(["%.10f"%p for p in params]))
            proc = subprocess.run("python %s %s" % (self.simulation_ref,args), stdout = subprocess.PIPE) # execute process
            bufferfile = proc.stdout.decode("utf-8") # get the output

            print("BUFFERFILE",bufferfile)
            data = np.load(bufferfile)   # read flat data
            os.unlink(bufferfile) # delete the temporary file after reading

            #print("DATA ITER",data)
            data = data.reshape(self.datashape) # re-shape data to original shape

            stop = time.time()
            print("%.2f" % (stop-start)," sec.; ", "DATA",data.shape, "Params", params)
            # print(stop-start)

            return data # return simulated data array

        elif self.mode == "callable":

            calble = self.simulation_ref
            data = calble(params)

            return data

    def calibrate(self, data, n_funcev, d_params):
        """
        EXPERIMENTAL and not tested!

        calibrate model parameters for a specific model

        :param data: dict of the form {name:array} with the calibration data
        :param n_funcev: number of function evaluations
        :param d_params: parameters for the distance measure
        """

        callable = self.method_dict[self.method] # find the calibration method to use
        callable(data_measured=data,bounds=self.param_bounds,n_funcev=n_funcev,d_params=d_params) # call the calibration method
        self.plot()

    def objective(self,pars,data,d_params):
            """
            EXPERIMENTAL and not tested!

            opimization objective

            :param  pars: parameters used to compute y~[t]
            :param data: observed data y[t]
            :pram d_params: parameters for distance measure d(y, y~)

            """

            data_measured = data
            self.data_measured = data_measured
            # print("data_simulated",pars)
            data_simulated = self.iter(pars)

            assert data_simulated.shape == self.datashape, "Simulated data has not the expected shape %s. It has shape %s" % (str(self.datashape),str(data_simulated.shape))

            dm =  self.dmeasure.d(data,data_simulated,pars,d_params)
            print("d = ", dm, "params", pars)

            return dm


    def plot_fit(self,xi,*args):
        """
        EXPERIMENTAL and not tested!
        """

        print("ITER")
        print("*args",*args)
        plt.figure()
        plt.style.use("macro")

        colors = ["black","red","blue","lime","magenta","orange"]
        for i in range(self.data_measured.shape[1]):
            plt.plot(self.data_measured[:,i],linestyle="dotted",color=colors[i%len(colors)],alpha=.8)

        simu = self.iter(xi)

        for i in range(self.data_measured.shape[1]):
            plt.plot(simu[:,i],linestyle="solid",color=colors[i%len(colors)])

        plt.xlabel("Time")
        plt.ylabel("Value")
        filename = "./figures/" + str(time.time())+ ".png"
        print("filename",filename)
        plt.savefig(filename)
        plt.close()


    def calibrate_scipy(self,data_measured,bounds,n_funcev,d_params,method):
        """
        EXPERIMENTAL and not tested!

        calibration step

        :param data: ndarray of shape (T,N) where T is the number of time steps and N is the number of different time series in the data
        :param bounds: ndarray of shape (P,2) where P is the number of parameters in the model
        :param n_funcev: desired number of function evaluations
        :param params: dict, further parameters of the calibration routine
        :param n_funcev: number of function evaluations
        """

        # 1) construct objective function
        #self.bjective(.):

        # 2) choose a starting point for parameters
        # self.x0

        # 3a) pre-minimize
        class MyBounds:
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html
            def __init__(self,xmax,xmin):
                self.xmax = xmax
                self.xmin = xmin

            def __call__(self,**kwagrs):
                x = kwargs["x_new"]
                tmax = bool(np.all(x <= self.xmax))
                tmin = bool(np.all(x >= self.xmin))
                print("accepted", tmax, tmin)
                return tmax and tmin

        my_bounds = MyBounds([b[1] for b in bounds], [b[0] for b in bounds])

        def printfun(x, f, accepted):
            print("at minimum %.4f accepted %d" % (f, int(accepted)))
            # sys.stdout.flush()

        # res = basinhopping(partial(self.objective,data=data), self.x0,niter = 1, T = 10, stepsize=1, disp = True, callback=printfun, accept_test = my_bounds)

        # rough pre-optimization to find a global optimum
        # x0=self.x0
        res = dual_annealing(partial(self.objective,d_params=d_params,data=data_measured), callback=self.plot_fit, bounds = bounds,x0=self.x0,maxiter=10000,maxfun=int(0.1*n_funcev)) # ,tol=0.1,workers=2,bounds=bounds, maxiter= 1)

        # 3) minimize!
        res = minimize(partial(self.objective,d_params=d_params,data=data_measured), res.x, callback=self.plot_fit,  bounds=bounds,method =method, options= {"maxiter": int(0.9*n_funcev)})

        # 4) return result
        self.plot_fit(res.x)

        # TODO maybe save a backup of res before plotting (?) in case the user shuts down the plot with the terminal or so

        print("OPTIMIZAITON RESULT", str(res))
        return res.x
